"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import scipy.misc as miscs
from scipy.special import binom
import cvx
from misc import bignom, Binomemoize
from cvxopt.lapack import syev
import numpy as np
import cvxopt as co
import math


def linear(X, norm=True):
	K = X.T * X
	
	if norm:
		for i in range(K.size[0]):
			if K[i,i] == 0:
				K[i,i] = 1.0
				
	return normalize(K) if norm else K
	

def normalize(K):
	"""
	@param K: the matrix
	@type K: cvxopt dense matrix
	@return: the row-normalized matrix
	@rtype: cvxopt dense matrix
	"""
	YY = cvx.diagonal_vec(K)
	YY = co.sqrt(YY)**(-1)
	return co.mul(K, YY*YY.T)

def force_normalize(K):
	for i in range(K.size[0]):
		if K[i,i] == 0:
			K[i,i] = 1.0
			
	return normalize(K)	

def spectral_complexity(K):
	a = cvx.trace(K)
	b = np.linalg.norm(K, "fro")
	return a / b


def spectral_complexity_norm(K):
	return (spectral_complexity(K) - 1.0) / (np.sqrt(K.size[0]) - 1.0)


def tanimoto(X, norm=False):
	d = co.matrix([sum(X[:,i].V) for i in xrange(X.size[1])])
	Yp = cvx.ones_vec(X.size[1]) * d.T
	Xp = d * cvx.ones_vec(X.size[1]).T
	Kl = X.T * X
	K = co.div(Kl, Xp + Yp - Kl)
	return normalize(K) if norm else K


def sparse_polynomial(X, c=1.0, d=2):
	K = X.T * X
	S = co.spmatrix(0.0, [], [], K.size)
	for i in range(1, d+1):
		cmb = miscs.comb(d, i)
		A = co.spmatrix(K.V**i, K.I, K.J)
		S += cmb * c**(d-i) * A
		
	return S
	

def polynomial(X, c=0.0, d=2.0, norm=True):
	K = ((X.T * X) + c)**d
	return normalize(K) if norm else K


def d_kernel(R, k, norm=True):
	
	n, m = R.size
	
	x_choose_k = [0]*(n+1)
	for i in range(1, n+1):
		x_choose_k[i] = binom(i,k)
	
	N = (2**k - 2) * x_choose_k[n]
	
	X = R.T*R
	U = co.matrix(1.0, (n,1))
	K = co.matrix(0.0, (X.size[0], X.size[1]))
	for i in range(m):
		#if (i+1) % 100 == 0:
		#	print "%d/%d" %(i+1,m)
		for j in range(i, m):
			r = int(((U - R[:,i]).T*((U - R[:,j])))[0])
			K[i,j] = K[j,i] = N + x_choose_k[int(X[i,j]) + r]
	
	return force_normalize(K) if norm else K
	
	
def md_kernel(R, k, norm=True):
	
	n, m = R.size
	
	x_choose_k = [0]*(n+1)
	for i in range(1, n+1):
		x_choose_k[i] = binom(i,k)
	
	nCk = x_choose_k[n]
	X = R.T*R
	
	K = co.matrix(0.0, (X.size[0], X.size[1]))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i, m):
			n_niCk = x_choose_k[n-int(X[i,i])]
			n_njCk = x_choose_k[n-int(X[j,j])]
			n_ni_nj_nijCk = x_choose_k[n+int(-X[i,i] - X[j,j] + X[i,j])]
			K[i,j] = K[j,i] = nCk - n_niCk - n_njCk + n_ni_nj_nijCk
				
	return force_normalize(K) if norm else K

def c_kernel(R, k, norm=True):
	n, m = R.size
	X = R.T*R
	U = co.matrix(1.0, (n,1))
	x_choose_k = [0]*(n+1)
	for i in range(1, n+1):
		x_choose_k[i] = binom(i,k)

	K = co.matrix(0.0, (m,m))
	for i in range(m):
		for j in range(i, m):
			r = int(((U - R[:,i]).T*((U - R[:,j])))[0])
			K[i,j] = K[j,i] = x_choose_k[int(X[i,j]) + r]
			
	return force_normalize(K) if norm else K


def mc_kernel(R, k, norm=True):
	n, m = R.size
	X = R.T*R

	x_choose_k = [0]*(n+1)
	for i in range(1, n+1):
		x_choose_k[i] = binom(i,k)

	K = co.matrix(0.0, (m,m))
	for i in range(m):
		for j in range(i, m):
			K[i,j] = K[j,i] = x_choose_k[int(X[i,j])]
			
	return force_normalize(K) if norm else K

def dnf_kernel(R, k, d, norm=True):

	n, m = R.size
	X = R.T*R
	
	C = binom(2**d*int(binom(n, d)), k) - 2*binom((2**d-1)*int(binom(n, d)), k)
	N = (2**d-2)*int(binom(n, d))
	
	U = co.matrix(1.0, (n,1))
	K = co.matrix(0.0, (m,m))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i, m):
			nij = int(X[i,j]) + ((U - R[:,i]).T*((U - R[:,j])))[0]
			K[i,j] = K[j,i] = C + binom(N + binom(nij, d), k) 
	
	return force_normalize(K) if norm else K
		
			
def mdnf_kernel(R, k, s, norm=True):
		
	n, m = R.size
	
	x_choose_s = {n : bignom(n,s)}
	nCs = x_choose_s[n]
	
	x_choose_k = {nCs : bignom(nCs,k)}
	a = x_choose_k[nCs]
	
	X = R.T*R
	
	if k == s == 1:
		K = X
	else:
		K = co.matrix(0.0, (m, m))
		for i in range(m):
			if (i+1) % 100 == 0:
				print "%d/%d" %(i+1,m)
			
			for j in range(i, m):
				
				xii = int(X[i,i])
				if xii not in x_choose_s:
					x_choose_s[xii] = bignom(xii, s)
				nCs_niCs = nCs - x_choose_s[xii]
				
				if nCs_niCs not in x_choose_k:
					x_choose_k[nCs_niCs] = bignom(nCs_niCs, k)
				b = x_choose_k[nCs_niCs]
				
				xjj = int(X[j,j])
				if xjj not in x_choose_s:
					x_choose_s[xjj] = bignom(xjj, s)
				nCs_njCs = nCs - x_choose_s[xjj]
				
				if nCs_njCs not in x_choose_k:
					x_choose_k[nCs_njCs] = bignom(nCs_njCs, k)
				c = x_choose_k[nCs_njCs]
				
				xij = int(X[i,j])
				if xij not in x_choose_s:
					x_choose_s[xij] = bignom(xij, s)
				nCs_niCs_njCs_nijCs = nCs - x_choose_s[xii] - x_choose_s[xjj] + x_choose_s[xij]
				
				if nCs_niCs_njCs_nijCs not in x_choose_k:
					x_choose_k[nCs_niCs_njCs_nijCs] = bignom(nCs_niCs_njCs_nijCs, k)
				d = x_choose_k[nCs_niCs_njCs_nijCs]
				
				K[i,j] = K[j,i] = float(a - c + d - b)
	
	return force_normalize(K) if norm else K
	
def mcnf_kernel(R, d, c, norm=True):
	n, m = R.size
		
	x_choose_d = {n : binom(n, d)}
	nCd = x_choose_d[n]
	
	X = R.T*R
	
	if c == d == 1:
		K = X
	else:
		K = co.matrix(0.0, (m, m))
		for i in range(m):
			if (i+1) % 100 == 0:
				print "%d/%d" %(i+1,m)
			
			for j in range(i, m):
					
				xii = n - int(X[i,i])
				if xii not in x_choose_d:
					x_choose_d[xii] = binom(xii, d)
				
				xjj = n - int(X[j,j])
				if xjj not in x_choose_d:
					x_choose_d[xjj] = binom(xjj, d)
				
				xij = n - int(X[i,i]) - int(X[j,j]) + int(X[i,j])
				if xij not in x_choose_d:
					x_choose_d[xij] = binom(xij, d)
				
				r = nCd - x_choose_d[xii] - x_choose_d[xjj] + x_choose_d[xij]
				K[i,j] = K[j,i] = binom(r, c)
	
	return force_normalize(K) if norm else K
	
	
def cnf_kernel(R, d, c, norm=True):
	n, m = R.size
		
	x_choose_k = [0]*(n+1)
	for i in range(1, n+1):
		x_choose_k[i] = binom(i, d)
	
	N = (2**d - 2) * x_choose_k[n]
	
	X = R.T*R
	U = co.matrix(1.0, (n,1))
	K = co.matrix(0.0, (X.size[0], X.size[1]))
	for i in range(m):
		#if (i+1) % 100 == 0:
		#	print "%d/%d" %(i+1,m)
		for j in range(i, m):
			r = int(((U - R[:,i]).T*((U - R[:,j])))[0])
			K[i,j] = K[j,i] = binom(N + x_choose_k[int(X[i,j]) + r], c)
	
	return force_normalize(K) if norm else K


def mdnf_kernel_old(R, norm=True):
	
	n, m = R.size
	X = R.T*R

	K = co.matrix(0.0, (m,m))
	for i in range(m):
		for j in range(i, m):
			K[i,j] = K[j,i] = 2.0**int(X[i,j]) - 1.0
			
	return normalize(K) if norm else K



#TESTING STUFF
def mdnf_kernel_norm(R):
	
	n, m = R.size
	X = R.T*R

	d = {i : mp.mpf(2**int(X[i,i]) - 1) for i in range(m)}
	
	K = co.matrix(0.0, (m,m))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i+1, m):
			K[i,j] = K[j,i] = float(mp.mpf(2**int(X[i,j]) - 1) / (mp.sqrt(d[i])*mp.sqrt(d[j])))
			
	for i in range(m):
		K[i,i] = 1.0	
		
	return K


'''
EXPERIMENTAL PART
'''

def k_over_d(k, d, bin, n, nij, ni_nij, nj_nij, n_ni_nj_nij):
	
	if k == 0:
		return bin.go(n, d)

	A1 = 0
	for s in range(k):
		A1 += bin.go(nij, s) * bin.go(ni_nij, k-s-1) * bin.go(nj_nij, k-s-1) * bin.go(n_ni_nj_nij, d+s-2*(k-1))
		  
	A2 = 0
	for s in range(k):
		D = 0
		for t in range(k-s, d-k+2):
			D += bin.go(ni_nij, t) * bin.go(n_ni_nj_nij, d-t-k+1)
		A2 += bin.go(nij, s) * bin.go(nj_nij, k-s-1) * D

	A3 = 0        
	for s in range(k):
		D = 0
		for t in range(k-s, d-k+2):
			D += bin.go(nj_nij, t) * bin.go(n_ni_nj_nij, d-t-k+1)
		A3 += bin.go(nij, s) * bin.go(ni_nij, k-s-1) * D

	return k_over_d(k-1, d, bin, n, nij, ni_nij, nj_nij, n_ni_nj_nij) - A1 - A2 - A3



def kd_kernel(R, d, k, norm=True):

	n, m = R.size
	X = R.T*R
	
	bin = Binomemoize()
	
	K = co.matrix(0.0, (m,m))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i, m):
			K[i,j] = K[j,i] = k_over_d(k, d, bin, n, int(X[i,j]), int(X[i,i] - X[i,j]), int(X[j,j] - X[i,j]), n+int(-X[i,i] - X[j,j] + X[i,j]))
			
	return force_normalize(K) if norm else K


def k_over_d_back(k, d, bin, n, nij, ni_nij, nj_nij, n_ni_nj_nij):
		
	if k == d:
		return bin.go(nij, d)

	A1 = 0
	for s in range(k+1):
		A1 += bin.go(nij, s) * bin.go(ni_nij, k-s) * bin.go(nj_nij, k-s) * bin.go(n_ni_nj_nij, d+s-2*k)
		  
	A2 = 0
	for s in range(k+1):
		D = 0
		for t in range(k-s+1, d-s+1):
			D += bin.go(ni_nij, t) * bin.go(n_ni_nj_nij, d-t-k)
		A2 += bin.go(nij, s) * bin.go(nj_nij, k-s) * D

	A3 = 0        
	for s in range(k+1):
		D = 0
		for t in range(k-s+1, d-s+1):
			D += bin.go(nj_nij, t) * bin.go(n_ni_nj_nij, d-t-k)
		A3 += bin.go(nij, s) * bin.go(ni_nij, k-s) * D

	return k_over_d_back(k+1, d, bin, n, nij, ni_nij, nj_nij, n_ni_nj_nij) + A1 + A2 + A3
		

def kd_kernel_back(R, d, k, norm=True):

	n, m = R.size
	X = R.T*R
	
	bin = Binomemoize()
	
	K = co.matrix(0.0, (m,m))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i, m):
			K[i,j] = K[j,i] = k_over_d_back(k, d, bin, n, int(X[i,j]), int(X[i,i] - X[i,j]), int(X[j,j] - X[i,j]), n+int(-X[i,i] - X[j,j] + X[i,j]))
			
	return force_normalize(K) if norm else K
	

def ekd_kernel(R, k, d, norm=True):
	n, m = R.size
	X = R.T*R
	
	bin = Binomemoize()
	K = co.matrix(0.0, (m,m))
	for i in range(m):
		if (i+1) % 100 == 0:
			print "%d/%d" %(i+1,m)
		for j in range(i, m):
			for s in range(k+1):
				K[i,j] += bin.go(int(X[i,j]), s) * bin.go(int(X[i,i] - X[i,j]), k-s) * bin.go(int(X[j,j] - X[i,j]), k-s) * bin.go(n-int(X[j,j])-int(X[i,i])+int(X[i,j]), d-2*k+s)
			K[j,i] = K[i,j]
	
	return force_normalize(K) if norm else K
	

def ekd_comb_kernel(R, d, norm=True):
	n, m = R.size
	K = co.matrix(0.0, (m,m))
	for k in range(1,d+1):
		K += ekd_kernel(R, k, d, False)
		
	return normalize(K)	if norm else K
	
'''
END OF EXPERIMENTAL PART
'''



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua

This module contains classes to read data sets from files.
By default, files are considered CSV (Comma Separated Values) with
a space as separator. Each line of the files are meant to be a record
with this form:

user_id || item_id || v :subscript:'1' || v :subscript:'2' || ... || v :subscript:'n',

where v :subscript:'i' are some kind of values and || are separators (spaces by default).
"""


class CSVReader(object):

	def __init__(self, filename, delimiter=' ', binarize=False):
		"""
		@param filename: the filename of the CSV file
		@type filename: string
		@param delimiter: the delimiter of the fields
		@type delimiter: string
		"""
		self.filename = filename
		self.delimiter = delimiter
		self.binarize = binarize

	# Reads the file and create the data set
	def read(self, data, implicit=False):
		"""
		@param data: the empty dataset
		@type data: BaseDataset
		@return: a dataset
		@rtype: UDataset
		"""
		with open(self.filename, 'r+') as f:
			contents = f.read().split("\n")
		
		for line in contents:
			r = line.strip().split(self.delimiter)
			if len(r) >= 2:
				(u, i, v) = self.interpret(r)
				if (u != -1):
					if implicit:
						data.add(u, i)
					else:
						data.add(u, i, v)
						


	# Interprets a single splitted line
	def interpret(self, args):
		"""
		@param args: strings' tuple
		@type args: tuple of strings 
		"""
		if len(args) == 2:
			return (int(args[0]), int(args[1]), Value(1.0))
		elif len(args) > 2:
			return (int(args[0]), int(args[1]), Value(1.0) if self.binarize else Value(float(args[2])))
		else: #ERROR
			return (-1, -1, None)


class MLReader(CSVReader):
	
	def read(self, data):
		with open(self.filename, 'r+') as f:
			contents = f.read().split("\n")
		
		for line in contents:
			r = line.strip().split(self.delimiter)
			if len(r) >= 2:
				(u, i, v) = self.interpret(r)
				if u != -1:
					data.add(u, i, v)

	
	def __init__(self, filename, delimiter=" "):
		super(MLReader, self).__init__(filename, delimiter)
	
	def interpret(self, args):
		if (len(args) > 3):
			v = Value(1.0, int(args[3])) if self.binarize else Value(float(args[2]), int(args[3]))
			return (int(args[0]), int(args[1]), v)
		else:
			return (int(args[0]), int(args[1]), Value(1.0))


class MSDReader(CSVReader):
	
	def __init__(self, filename, delimiter="\t"):
		super(MSDReader, self).__init__(filename, delimiter)
	
	def interpret(self, args):
		return (args[0], args[1], Value(1.0))


# Represents a generic value for the ratings
class Value(object):
	
	def __init__(self, v, t=0):
		"""
		@param v: a generic value
		@type v: float
		@param t: the timestamp (default: 0)
		@type t: int 
		"""
		self.val = v
		self.time = t
	
	# Returns the value
	def get_float(self):
		return self.val
	
	# Returns the timestamp
	def get_time(self):
		return self.time

	def __eq__(self, obj):
		return self.__dict__ == obj.__dict__
	
	def __hash__(self):
		return hash((self.val, self.time))

	def __repr__(self):
		if self.time != 0:
			return "(%s, %s)" %(self.val, self.time)
		else:
			return "%s" %self.val 



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import cvxopt as co
from scipy import sparse
from mapping import IdentityMapping
from reader import Value
import numpy as np


# Abstract RS dataset
class BaseDataset(object):

	def __init__(self,
				user_mapping=IdentityMapping(),
				item_mapping=IdentityMapping()):
		"""
		@param user_mapping: map users id to the integers interval [0, N]
		@type user_mapping: dictionary {any_immutable_type : int}
		@param item_mapping: map items id to the integers interval [0, N]
		@type item_mapping: dictionary {any_immutable_type : int}
		"""
		self.user_mapping = user_mapping
		self.item_mapping = item_mapping
		self.users = set()
		self.items = set()

		self.data = {}

	# Adds a new pair (user, item) to the dataset
	def add(self, user, item):
		"""
		@param user: user id
		@param item: item id
		@return: tuple of new user id and new item id
		@rtype: tuple (int, int)
		"""
		u = self.user_mapping.add(user)
		i = self.item_mapping.add(item)

		self.users.add(u)
		self.items.add(i)

		return u, i

	# Returns the number of users
	def num_users(self):
		"""
		@return: the number of users
		@rtype: int
		"""
		return len(self.users)

	# Returns the number of items
	def num_items(self):
		"""
		@return: the number of items
		@rtype: int
		"""
		return len(self.items)

	# Returns the number of ratings
	def num_ratings(self):
		"""
		@return: the number of users
		@rtype: int
		"""
		raise NotImplementedError

	# Returns the string representation of the dataset
	def __repr__(self):
		"""
		@return: the string representation of the dataset
		@rtype: string
		"""
		return "Data-set with %d users and %d items." %(self.num_users(), self.num_items())

	# Returns the set of items rated by the given user
	def get_items(self, user):
		"""
		@param user: the user
		@type user: int
		@return: set of items rated by user
		@rtype: set of int
		"""
		raise NotImplementedError
	
	# Returns the numpy dense rating matrix corresponding to the dataset
	def to_numpy_matrix(self):
		"""
		@return: the numpy dense rating matrix corresponding to the dataset
		@rtype: numpy dense matrix
		"""
		return NotImplementedError

	# Returns the numpy sparse rating matrix corresponding to the dataset
	def to_numpy_sparse_matrix(self):
		"""
		@return: the numpy dense rating matrix corresponding to the dataset
		@rtype: numpy sparse matrix
		"""
		raise NotImplementedError

	# Returns the cvxopt dense rating matrix corresponding to the dataset
	def to_cvxopt_matrix(self):
		"""
		@return: the cvxopt dense rating matrix corresponding to the dataset
		@rtype: cvxopt dense matrix
		"""
		return NotImplementedError

	# Returns the cvxopt sparse rating matrix corresponding to the dataset
	def to_cvxopt_sparse_matrix(self):
		"""
		@return: the cvxopt sparse rating matrix corresponding to the dataset
		@rtype: cvxopt sparse matrix
		"""
		raise NotImplementedError
	
	# Returns the cvxopt rating matrix corresponding to the dataset
	def to_cvxopt(self, spr=False):
		"""
		@param spr: whether the representation is sparse or not
		@type spr: boolean
		@return: the cvxopt rating matrix corresponding to the dataset
		@rtype: cvxopt matrix
		"""
		return self.to_cvxopt_matrix() if not spr else self.to_cvxopt_sparse_matrix()
	
	# Returns the numpy rating matrix corresponding to the dataset
	def to_numpy(self, spr=False):
		"""
		@param spr: whether the representation is sparse or not
		@type spr: boolean
		@return: the numpy rating matrix corresponding to the dataset
		@rtype: numpy matrix
		"""
		return self.to_numpy_matrix() if not spr else self.to_numpy_sparse_matrix()
	

#Generic RS dataset
class Dataset(BaseDataset):

	def __init__(self,
				user_mapping=IdentityMapping(),
				item_mapping=IdentityMapping()):
		super(Dataset, self).__init__(user_mapping, item_mapping)
		self.scale = Scale()

	def add(self, user, item, rating=Value(0.0)):
		(u, i) = super(Dataset, self).add(user, item)
		self.scale.add(rating.get_float())
		self.data[(u, i)] = rating

		return u, i

	def num_ratings(self):
		return len(self.data)

	def get_items(self, user):
		return set([i for u,i in self.data if u == user])

	def to_numpy_matrix(self):
		result = np.matrix(0.0, (self.num_users(), self.num_items()))
		for (u, i), v in self.data.iteritems():
			result[u, i] = v.get_float()
		return result

	def to_numpy_sparse_matrix(self):
		rows, cols, values = [],[],[]
		for (u, i), v in self.data.iteritems():
			rows += [u]
			cols += [i]
			values += [v.get_float()]

		return sparse.coo_matrix((values, (rows, cols)), shape=(self.num_users(), self.num_items()))

	def to_cvxopt_matrix(self):
		result = co.matrix(0.0, (self.num_users(), self.num_items()))
		for (u, i), v in self.data.iteritems():
			result[u, i] = v.get_float()
		return result
	
	def to_cvxopt_sparse_matrix(self):
		rows, cols, values = [], [], []
		for (u, i), v in self.data.iteritems():
			rows += [u]
			cols += [i]
			values += [v.get_float()]

		return co.spmatrix(values, rows, cols, (self.num_users(), self.num_items()))


#User-centered RS dataset
class UDataset(BaseDataset):

	def __init__(self,
				user_mapping=IdentityMapping(),
				item_mapping=IdentityMapping()):
		super(UDataset, self).__init__(user_mapping, item_mapping)
		self.scale = Scale()
		self.count = 0

	def add(self, user, item, rating=Value(1.0)):
		(u, i) = super(UDataset, self).add(user, item)
		self.scale.add(rating.get_float())
		if u not in self.data:
			self.data[u] = set()
		self.data[u].add((i, rating))
		self.count += 1

		return u, i

	def num_ratings(self):
		return self.count

	def get_items(self, user):
		return set([i for (i,_) in self.data[user]]) if user in self.data else set()

	def to_numpy_matrix(self):
		result = np.zeros((self.num_users(), self.num_items()))
		for u, s in self.data.iteritems():
			for (i, v) in s:
				result[u, i] = v.get_float()
		return result

	def to_numpy_sparse_matrix(self):
		rows, cols, values = [],[],[]
		for u, s in self.data.iteritems():
			for (i, v) in s:
				rows += [u]
				cols += [i]
				values += [v.get_float()]

		return sparse.coo_matrix((values, (rows, cols)), shape=(self.num_users(), self.num_items()))

	def to_cvxopt_matrix(self):
		result = co.matrix(0.0, (self.num_users(), self.num_items()))
		for u, s in self.data.iteritems():
			for (i, v) in s:
				result[u, i] = v.get_float()
		return result
	
	def to_cvxopt_sparse_matrix(self):
		rows, cols, values = [], [], []
		for u, s in self.data.iteritems():
			for (i, v) in s:
				rows += [u]
				cols += [i]
				values += [v.get_float()]
		
		return co.spmatrix(values, rows, cols, (self.num_users(), self.num_items()))


#Item-centered RS dataset
class IDataset(BaseDataset):

	def __init__(self,
				user_mapping=IdentityMapping(),
				item_mapping=IdentityMapping()):
		super(IDataset, self).__init__(user_mapping, item_mapping)
		self.scale = Scale()
		self.count = 0

	def add(self, user, item, rating=Value(1.0)):
		(u, i) = super(IDataset, self).add(user, item)
		self.scale.add(rating.get_float())
		if i not in self.data:
			self.data[i] = set()
		self.data[i].add((u, rating))
		self.count += 1

		return u, i

	def num_ratings(self):
		return self.count
		
	def get_users(self, item):
		return set([u for (u,_) in self.data[item]]) if item in self.data else set()

	def to_numpy_matrix(self):
		result = np.zeros((self.num_users(), self.num_items()))
		for i, s in self.data.iteritems():
			for (u, v) in s:
				result[u, i] = v.get_float()
		return result

	def to_numpy_sparse_matrix(self):
		rows, cols, values = [],[],[]
		for i, s in self.data.iteritems():
			for (u, v) in s:
				rows += [u]
				cols += [i]
				values += [v.get_float()]

		return sparse.coo_matrix((values, (rows, cols)), shape=(self.num_users(), self.num_items()))

	def to_cvxopt_matrix(self):
		result = co.matrix(0.0, (self.num_users(), self.num_items()))
		for i, s in self.data.iteritems():
			for (u, v) in s:
				result[u, i] = v.get_float()
		return result
	
	def to_cvxopt_sparse_matrix(self):
		rows, cols, values = [], [], []
		for i, s in self.data.iteritems():
			for (u, v) in s:
				rows += [u]
				cols += [i]
				values += [v.get_float()]
		
		return co.spmatrix(values, rows, cols, (self.num_users(), self.num_items()))


# Class which represents a scale of values
class Scale(object):

	def __init__(self):
		self.values = set()
		self.min = float("+inf")
		self.max = float("-inf")

	# Adds a value to the scale
	def add(self, value):
		"""
		@param value: the value to add to the scale
		@type value: number
		"""
		self.values.add(value)

		if value > self.max:
			self.max = value
		elif value < self.min:
			self.min = value

#TODO data-set statistics




"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

from recommender import RecEngine
import baseline as base
import pyros.utils.cvx as utc
import cvxopt as co
import cvxopt.solvers as solver
import numpy as np
from scipy.sparse import csr_matrix
from pyros.utils.misc import fast_sparse_matrix

'''
Matrix-based implementation of the algorithm described
in "Efficient Top-N Recommendation for Very Large Scale Binary Rated Datasets"
by F. Aiolli
'''
class I2I_Asym_Cos(RecEngine):
	def __init__(self, data, alpha=0.5, q=1.0):
		super(I2I_Asym_Cos, self).__init__(data)
		self.ratings = self.data.to_cvxopt(True)
		self.model = None
		self.alpha = alpha
		self.locality = q

	def train(self, test_users=None):
		self.model = self.ratings.T * self.ratings
		diag = utc.diagonal_vec(self.model)
		row_mat = diag * utc.ones_vec(self.n_items).T
		col_mat = utc.ones_vec(self.n_items) * diag.T

		row_mat **= self.alpha
		col_mat **= 1 - self.alpha
		self.model = co.div(self.model, co.mul(row_mat, col_mat))
		self.model = self.model**self.locality

	def get_scores(self, u):
		return self.model * self.ratings[u, :].T
	
	def get_params(self):
		return {"alpha" : self.alpha, "locality" : self.locality}
	

'''
Implementation of the algorithm described in
"Convex AUC Optimization for Top-N Recommendation with Implicit Feedback"
by F. Aiolli
'''
class CF_OMD(RecEngine):
	def __init__(self, data, lp=1.0, ln=1000.0, spr=False):
		super(CF_OMD, self).__init__(data)
		self.lambda_p = lp
		self.lambda_n = ln
		self.X = utc.normalize_cols(self.data.to_cvxopt(spr))
		self.model = None
		self.sol = {}
		
		
	def train(self, test_users=None):
		self.model = {}
		iset = set(range(self.n_items))
		
		if test_users is None:
			test_users = range(self.n_users)
		
		for i, u in enumerate(test_users):
			#if (i+1) % 100 == 0:
			print "%d/%d" %(i+1, len(test_users))

			Xp_set = self.data.get_items(u)
			Xn_set = iset - Xp_set

			Z = co.spdiag([1.0 if i in Xp_set else -1.0 for i in iset])
			K = 2 * (Z * self.X.T * self.X * Z)
			I = co.spdiag([self.lambda_p if i in Xp_set else self.lambda_n for i in iset])
			P = K + I

			o = utc.zeroes_vec(self.n_items)
			G = -utc.identity(self.n_items)
			A = co.matrix([[1.0 if i in Xp_set else 0.0 for i in iset],
						   [1.0 if j in Xn_set else 0.0 for j in iset]]).T
			b = co.matrix([1.0, 1.0])
			P = co.sparse(P)

			solver.options['show_progress'] = False
			sol = solver.qp(P, o, G, o, A, b)
			self.sol[u] = sol
			self.model[u] = self.X.T * self.X * Z * sol['x']

		# endfor
		return self
	
	def get_params(self):
		return {"lambda_p" : self.lambda_p, "lambda_n" : self.lambda_n}
	
	def get_scores(self, u):
		return self.model[u]


'''
Implementation of the algorithm (which is a simplification of CF-OMD) described in
"Kernel based collaborative filtering for very large scale top-N item recommendation"
by M.Polato and F. Aiolli
'''
class ECF_OMD(RecEngine):
	def __init__(self, data, lp=0.1, spr=False):
		super(ECF_OMD, self).__init__(data)
		self.lambda_p = lp
		self.X = utc.normalize_cols(self.data.to_cvxopt(spr)).T
		self.Xn_ave = utc.ones_vec(self.n_items).T * self.X
		self.model = None
		self.sol = {}
		

	def train(self, test_users=None):
		self.model = {}

		if test_users is None:
			test_users = range(self.n_users)
		
		for i, u in enumerate(test_users):
			if (i+1) % 100 == 0:
				print "%d/%d" %(i+1, len(test_users))

			Xp_list = list(self.data.get_items(u))
			
			np = len(Xp_list)
			nn = float(self.n_items - np)
			
			Xp = co.matrix(self.X[Xp_list, :])
			Xp_ave = utc.ones_vec(np).T * Xp
			Xn_ave_u = (self.Xn_ave - Xp_ave) / nn
			
			kn = Xp * Xn_ave_u.T
			K = Xp * Xp.T
			I = self.lambda_p * utc.identity(np)
			P = K + I
			q = -kn
			G = -utc.identity(np)
			h = utc.zeroes_vec(np)
			A = utc.ones_vec(np).T
			b = co.matrix(1.0)

			solver.options['show_progress'] = False
			sol = solver.qp(P, q, G, h, A, b)
			self.sol[u] = co.matrix(1.0/(self.n_items - np), (self.n_items, 1))
			self.sol[u][Xp_list] = sol['x']
			self.model[u] = self.X * (Xp.T * sol['x'] - Xn_ave_u.T) # not normalized

		# endfor
		return self

	def get_params(self):
		return {"lambda_p" : self.lambda_p}

	def get_scores(self, u):
		return  self.model[u]


'''
Implementation of the algorithm (which is a "kernelification" of ECF-OMD) described in
"Kernel based collaborative filtering for very large scale top-N item recommendation"
by M.Polato and F. Aiolli
'''
class CF_KOMD(RecEngine):
	def __init__(self, data, K=None, lp=0.1, spr=False):
		super(CF_KOMD, self).__init__(data)
		self.lambda_p = lp
		self.K = K
		self.q_ = co.matrix(0.0, (self.n_items, 1))
		for i in xrange(self.n_items):
			self.q_[i,0] = sum(self.K[i,:]) / float(self.n_items) #-1
		
		self.model = None
		self.sol = {} #TODO
		
		
	def train(self, test_users=None):
		self.model = {}
		
		if test_users is None:
			test_users = range(self.n_users)
		
		for i, u in enumerate(test_users):
			#if (i+1) % 100 == 0:
			#	print "%d/%d" %(i+1, len(test_users))

			Xp = list(self.data.get_items(u))
			np = len(Xp)
			
			kp = self.K[Xp, Xp]
			kn = self.q_[Xp,:]
			
			I = self.lambda_p * utc.identity(np)
			P = kp + I
			q = -kn
			G = -utc.identity(np)
			h = utc.zeroes_vec(np)
			A = utc.ones_vec(np).T
			b = co.matrix(1.0)

			solver.options['show_progress'] = False
			sol = solver.qp(P, q, G, h, A, b)
			
			self.model[u] = self.K[Xp,:].T * sol['x'] - self.q_
			#print self.model[u]

		# endfor
		return self

	def get_params(self):
		return {"lambda_p" : self.lambda_p}

	def get_scores(self, u):
		return self.model[u]


class CF_KOMD_CS_POP(CF_KOMD):
	def __init__(self, data, K=None, lp=0.1, spr=False):
		super(CF_KOMD_CS_POP, self).__init__(data, K, lp, spr)
		self.pop_model = base.Popular(data)
	
	def train(self, test_users=None):
		self.pop_model.train()
		super(self.__class__, self).train(test_users & self.data.users)
	
	def get_scores(self, u):
		if u in self.model:
			return self.model[u]
		else:
			return self.pop_model.get_scores(u)

	

'''
Implementation of the algorithm SLIM described in
"SLIM: Sparse Linear Methods for Top-N Recommender Systems"
by Xia Ning and George Karypis
'''
class SLIM(RecEngine):
	
	def __init__(self, data, beta=4, lam=.5):
		super(self.__class__, self).__init__(data)
		self.R = self.data.to_cvxopt(False)
		self.beta = beta
		self._lambda = lam
		self.model = None
	
	'''
	Load the model from a .mat file, which is the type of output given by the
	SLIM implementation provided by the authors (https://www-users.cs.umn.edu/~ningx005/slim/html/)
	'''
	def load_model(self, filename):
		f = open(filename, "rb")
		self.model = {}
		content = f.readlines()
		for u, line in enumerate(content):
			#if (u+1) % 500 == 0:
			#	print "%d/%d" %(u+1, len(content))
			uss = line.strip().split(" ")
			ss = co.matrix(0.,(1,self.n_users))
			if len(uss) > 1:
				for x in range(0,len(uss),2):
					ss[0,int(uss[x][0])-1] = float(uss[x+1])
			self.model[u] = self.R.T * ss.T
	
	def train(self, test_users=None):
		self.model = {}
		
		if test_users is None:
			test_users = range(self.n_users)
		
		RR = self.R*self.R.T 
		P = RR + self.beta * utc.identity(self.n_users) 
		G = -utc.identity(self.n_users)
		h = utc.zeroes_vec(self.n_users)
		b = utc.zeroes_vec(1)		
		
		for i, u in enumerate(test_users):
			#if (i+1) % 100 == 0:
			#print "%d/%d" %(i+1, len(test_users))
			q = self._lambda * utc.ones_vec(self.n_users) - RR[u,:].T 
			A = co.matrix(0.,(1,self.n_users))
			A[0,u] = 1.

			solver.options['show_progress'] = False
			sol = solver.qp(P, q, G, h, A, b)
			
			self.model[u] = self.R.T * sol['x']
		
		return self
		

	def get_params(self):
		return {"lambda" : self._lambda, "beta" : self.beta}

	def get_scores(self, u):
		return  self.model[u]


'''
Implementation of the algorithm WRMF described in
"Collaborative Filtering for Implicit Feedback Datasets"
by Yifan Hu, Yehuda Koren and Chris Volinsky
and in 
"One-Class Collaborative Filtering"
by Rong Pan, Yunhong Zhou, Bin Cao, Nathan N. Liu, Rajan Lukose, Martin Scholz and Qiang Yang
'''
class WRMF(RecEngine):
	'''
	latent_factors : int
		Number of latent factors.
	alpha : float
		Confidence weight, confidence c = 1 + alpha*r where r is the observed "rating".
	lbda : float
		Regularization constant.
	num_iters : int
		Number of iterations of alternating least squares.
	'''
	def __init__(self, data, latent_factors, alpha=0.01, lbda=0.015, num_iters=15):
		super(self.__class__, self).__init__(data)
		self.R = fast_sparse_matrix(self.data.to_numpy_sparse_matrix().tocsr())
		self.latent_factors = latent_factors
		self.alpha = alpha
		self.lbda = lbda
		self.num_iters = num_iters
		self.model = None
		
	def init_factors(self,num_factors,assign_values=True):
		if assign_values:
			return self.latent_factors**-0.5*np.random.random_sample((num_factors,self.latent_factors))
		return np.empty((num_factors,self.latent_factors))

	def train(self, test_users=None):
		num_users,num_items = self.R.shape

		self.U = self.init_factors(num_users, False)  # don't need values, will compute them
		self.V = self.init_factors(num_items)
		for it in xrange(self.num_iters):
			print 'iteration',it
			# fit user factors
			VV = self.V.T.dot(self.V)
			for u in xrange(num_users):
				# get (positive i.e. non-zero scored) items for user
				indices = self.R.X[u].nonzero()[1]
				if indices.size:
					self.U[u,:] = self.update(indices,self.V,VV)
				else:
					self.U[u,:] = np.zeros(self.latent_factors)
			# fit item factors
			UU = self.U.T.dot(self.U)
			for i in xrange(num_items):
				indices = self.R.fast_get_col(i).nonzero()[0]
				if indices.size:
					self.V[i,:] = self.update(indices,self.U,UU)
				else:
					self.V[i,:] = np.zeros(self.latent_factors)
		
		self.model = np.dot(self.U,self.V.T)
		return self

	def update(self,indices,H,HH):
		Hix = H[indices,:]
		M = HH + self.alpha*Hix.T.dot(Hix) + np.diag(self.lbda*np.ones(self.latent_factors))
		return np.dot(np.linalg.inv(M),(1+self.alpha)*Hix.sum(axis=0))
	
	def get_params(self):
		return {'latent_factors':self.latent_factors, 'alpha':self.alpha, 'lambda':self.lbda, 'num_iters':self.num_iters}
	
	def get_scores(self, u):
		return co.matrix(self.model[u])
	

'''
Implementation of the algorithm BPRMF described in
"BPR: Bayesian Personalized Ranking from Implicit Feedback"
by Steffen Rendle, Christoph Freudenthaler, Zeno Gantner and Lars Schmidt-Thieme
'''
class BPRMF(RecEngine):
	def __init__(self, data, factors=40, learn_rate=0.05, num_iters=15, 
				 init_mean=0.1, init_stdev=0.1, reg_u=0.0025, reg_i=0.0025, reg_bias=0):
		super(self.__class__, self).__init__(data)

		self.factors = factors
		self.learn_rate = learn_rate
		self.init_mean = init_mean
		self.init_stdev = init_stdev
		self.num_iters = num_iters 
		self.reg_bias = reg_bias
		self.reg_u = reg_u
		self.reg_i = reg_i
		self.reg_j = reg_i
		
		# internal vars
		self.loss = None
		self.loss_sample = list()
		self._create_factors()

	def _create_factors(self):
		self.p = np.random.normal(self.init_mean, self.init_stdev, (self.data.num_users(), self.factors))
		self.q = np.random.normal(self.init_mean, self.init_stdev, (self.data.num_items(), self.factors))
		# self.bias = self.init_mean * np.random.randn(self.number_items, 1) + self.init_stdev ** 2
		self.bias = np.zeros(self.data.num_items(), np.double)

	def _sample_triple(self):
		user = np.random.choice(list(self.data.users))
		user_items = list(self.data.get_items(user))
		item_i = np.random.choice(user_items)
		item_j = np.random.choice(list(self.data.items - self.data.get_items(user)))
		return user, item_i, item_j

	def _predict(self, user, item):
		return self.bias[item] + np.dot(self.p[user], self.q[item])

	def _update_factors(self, user, item_i, item_j):
		# Compute Difference
		eps = 1 / (1 + np.exp(self._predict(user, item_i) - self._predict(user, item_j)))

		self.bias[item_i] += self.learn_rate * (eps - self.reg_bias * self.bias[item_i])
		self.bias[item_j] += self.learn_rate * (eps - self.reg_bias * self.bias[item_j])

		# Adjust the factors
		u_f = self.p[user]
		i_f = self.q[item_i]
		j_f = self.q[item_j]

		# Compute and apply factor updates
		self.p[user] += self.learn_rate * ((i_f - j_f) * eps - self.reg_u * u_f)
		self.q[item_i] += self.learn_rate * (u_f * eps - self.reg_i * i_f)
		self.q[item_j] += self.learn_rate * (-u_f * eps - self.reg_j * j_f)

	def _compute_loss(self):
		ranking_loss = 0
		for sample in self.loss_sample:
			x_uij = self._predict(sample[0], sample[1]) - self._predict(sample[0], sample[2])
			ranking_loss += 1 / (1 + np.exp(x_uij))

		complexity = 0
		for sample in self.loss_sample:
			complexity += self.reg_u * np.linalg.norm(self.p[sample[0]]) ** 2
			complexity += self.reg_i * np.linalg.norm(self.q[sample[1]]) ** 2
			complexity += self.reg_j * np.linalg.norm(self.q[sample[2]]) ** 2
			complexity += self.reg_bias * self.bias[sample[1]] ** 2
			complexity += self.reg_bias * self.bias[sample[2]] ** 2

		return ranking_loss + 0.5 * complexity

	# Perform one iteration of stochastic gradient ascent over the training data
	# One iteration is samples number of positive entries in the training matrix times
	def train(self, test_users=None):
		num_sample_triples = int(np.sqrt(len(self.data.users)) * 100)
		for _ in range(num_sample_triples):
			self.loss_sample.append(self._sample_triple())
		self.loss = self._compute_loss()

		for n in range(self.num_iters):
			for _ in range(self.data.num_ratings()):
				u, i, j = self._sample_triple()
				self._update_factors(u, i, j)

			actual_loss = self._compute_loss()
			if actual_loss > self.loss:
				self.learn_rate *= 0.5
			elif actual_loss < self.loss:
				self.learn_rate *= 1.1
			self.loss = actual_loss

			print "step::", n, "RMSE::", self.loss, "lrate::", self.learn_rate
		return self

	def get_params(self):
		return {"factors": self.factors, "learn_rate": self.learn_rate, "": self.num_iters, "reg_user" : self.reg_u, "reg_item": self.reg_i, "reg_bias" : self.reg_bias}

	def get_scores(self, u):
		model_u = np.dot(self.p[[u]], self.q.T) + self.bias
		return co.matrix(model_u.T)




"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import time
import os
import pickle
import cvx as utc
import numpy as np
from sklearn.datasets import dump_svmlight_file


def load_dataset(reader):
	uci = reader.read()
	R = uci.to_cvxopt_sparse_matrix()
	X = utc.normalize_cols_sparse(R)
	
	return X, R, uci


# Saves the object to file
def save_to_file(obj, fname):
	"""
	@param obj: the object to save
	@param fname: the file name
	@type fname: string
	"""
	pickle.dump(obj, open(fname, "wb"))


# Loads object from file
def load_from_file(fname):
	"""
	@param fname: the file name
	@type fname: string
	@return: the object
	"""
	return pickle.load(open(fname, "rb"))


# Saves results to file
def save_results(path, rec, result):
	with open(path, "w") as f:
		f.write(rec.get_fullname() + "\n")
		f.write(result)


# Sorts the list by value
def sort(l, skip, dec=True):
	"""
	@param l: the list to sort
	@type l: list of numbers
	@param skip: list of indexes to skip
	@type skip: list of int
	@param dec: whether to sort in descending order or not
	@type dec: boolean (default: True)
	"""
	d = {k:l[k] for k in xrange(len(l)) if k not in skip}
	return sorted(d.keys(),key=lambda s:d[s],reverse=dec)
	
	
# Calculates the time taken by the given method/function
def timing(method):
	"""
	@param method: the method to time
	@type method: function closure
	"""
	def timed(*args, **kw):
		ts = time.time()
		result = method(*args, **kw)
		te = time.time()

		#print '%r (%r, %r) %2.2f sec' % \
		#(method.__name__, args, kw, te-ts)
		print '%r %r %2.2f sec' %(method.__name__, args, te-ts)
		#logging.debug('%r %r %2.2f sec' %(method.__name__, args, te-ts))
		return result
	return timed

def save_as_svmlight(M, yid, fname):
	dump_svmlight_file(M[:,[x for x in range(M.shape[1]) if x != yid]], M[:,yid], fname)


class Binomemoize():
	def __init__(self):
		self.mem = {}
	
	def go(self, n, k):
		if n in self.mem:
			if k not in self.mem[n]:
				self.mem[n][k] = binom(n, k)
		else:
			self.mem[n] = {k : binom(n, k)}
		
		return self.mem[n][k]

def bignom(n,k):
	if n < k: return mp.mpf(0.0)
	res = mp.mpf(1.0)
	for i in range(0,k):
		res = res * mp.mpf(float(n-i)/(k-i))
	return res


class fast_sparse_matrix():
	def __init__(self,X,col_view=None):
		self.X = X.tocsr()
		if col_view is not None:
			self.col_view = col_view
		else:
			# create the columnar index matrix
			ind = self.X.copy()
			ind.data = np.arange(self.X.nnz)
			self.col_view = ind.tocsc()

	@property
	def shape(self):
		return self.X.shape

	def fast_get_col(self,j):
		col = self.col_view[:,j].copy()
		col.data = self.X.data[col.data]
		return col



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua

This module contains utility methods for manipulating cvxopt objects.
"""

import cvxopt as co
from math import sqrt


# Creates a n-dimensional vector initialized with all zeroes
def zeroes_vec(n):
	"""
	@param n: the dimension of the vector
	@type n: int
	@return: a column n-dimensional zero vector 
	@rtype: cvxopt dense matrix
	"""
	return co.matrix(0.0, (n,1))


# Creates a n-dimensional vector initialized with all ones
def ones_vec(n):
	"""
	@param n: the dimension of the vector
	@type n: int
	@return: a column n-dimensional one vector 
	@rtype: cvxopt dense matrix
	"""
	return co.matrix(1.0, (n,1))


# Returns the diagonal of the given matrix
def diagonal_vec(X):
	"""
	@param X: dense square matrix
	@type X: cvxopt dense matrix
	@return: a column vector with the diagonal elements of X
	@rtype: cvxopt dense matrix
	"""
	return co.matrix([X[i,i] for i in xrange(X.size[0])])


# Creates a diagonal matrix using the given vector as diagonal
def diag(v):
	"""
	@param v: the vector to put in the diagonal
	@type v: cvxopt dense matrix
	@return: diagonal matrix
	@rtype: cvxopt dense matrix
	"""
	result = co.matrix(0.0, (v.size[0], v.size[0]))
	for i, x in enumerate(v):
		result[i, i] = x
	return result


# Creates the identity matrix of the given dimension
def identity(n):
	"""
	@param n: the dimension of the matrix (n by n)
	@type n: int
	@return: the identity matrix
	@rtype: cvxopt dense matrix
	"""
	return diag(ones_vec(n))


# Calculates the trace of the matrix
def trace(K):
	"""
	@param K: square matrix
	@type K: cvxopt dense matrix
	@return: the trace of the matrix
	@rtype: float
	"""
	return sum(diagonal_vec(K))


# Normalizes the vector
def normalize_vec(v):
	"""
	@param v: the vector
	@type v: cvxopt dense matrix
	@return: the normalized vector
	@rtype: cvxopt dense matrix
	"""
	return v / sqrt(sum(v**2))


# Divides the rows for the sum of its elements
def normalize_rows(X):
	"""
	@param X: the matrix
	@type X: cvxopt dense matrix
	@return: the row-normalized matrix
	@rtype: cvxopt dense matrix
	"""
	d = diagonal_vec(X*X.T)
	N = co.sqrt(d * ones_vec(X.size[1]).T)
	return co.div(X,N)


# Divides the rows for the sum of its elements
def normalize_cols(X):
	"""
	@param X: the matrix
	@type X: cvxopt dense matrix
	@return: the col-normalized matrix
	@rtype: cvxopt dense matrix
	"""
	d = diagonal_vec(X.T*X)
	N = co.sqrt(ones_vec(X.size[0])*d.T)
	return co.div(X,N)

# Divides the columns for the sum of its elements
def normalize_cols_sparse(X):
	"""
	@param X: the matrix
	@type X: cvxopt matrix
	@return: the col-normalized matrix
	@rtype: cvxopt sparse matrix
	"""
	d = co.matrix([sum(X[:,i].V) for i in xrange(X.size[1])])
	N = co.sqrt(d.T)
	I, J = X.I, X.J
	V = []
	for j in X.J:
		V += [(1.0 / N[j]) if N[j] != 0.0 else 0.0]
	return co.spmatrix(V, I, J)

#def min_enclosing_ball(X):
	

# Sorts the vector by value
def sort(v, skip, dec=True):
	"""
	@param v: the vector to sort
	@type v: cvxopt dense matrix
	@param skip: vector of indexes to skip
	@type skip: cvxopt dense matrix
	@param dec: whether to sort in descending order or not
	@type dec: boolean (default: True)
	@return: ordered list by value
	@rtype: list
	"""
	d = {k:v[k] for k in xrange(v.size[0]) if k not in skip}
	return sorted(d.keys(),key=lambda s:d[s],reverse=dec)


# Applies the sigmoid function to the matrix
def sigmoid(X):
	return (1.0 + co.exp(-X))**(-1)

# Calculates the density of the matrix
def density(X):
	z = 0
	for i in range(X.size[0]):
		for j in range(X.size[1]):
			z += X[i,j] > 0.0
	
	return float(z) / (X.size[0] * X.size[1])


"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import cvxopt as co
import pyros.utils.cvx as ut
from recommender import RecEngine


class Random(RecEngine):
	def __init__(self, data):
		super(self.__class__, self).__init__(data)
		self.rand = None

	def train(self):
		self.rand = co.uniform(self.n_users, self.n_items)
		return self

	def get_scores(self, u):
		return self.rand[u, :]


class Popular(RecEngine):
	def __init__(self, data):
		super(self.__class__, self).__init__(data)
		self.popular = None

	def train(self):
		ratings = self.data.to_cvxopt_matrix()
		self.popular = (ut.ones_vec(self.n_users).T * ratings).T
		return self

	def get_scores(self, u):
		return self.popular


class Constant(RecEngine):
	def __init__(self, data, k=None):
		super(self.__class__, self).__init__(data)
		if k is None:
			self.constant = (self.data.scale.max - self.data.scale.min) / 2.0
		else:
			self.constant = k
	
	def get_params(self):
		return {"constant" : self.constant}
	
	def get_scores(self, u):
		return co.matrix(self.constant, (self.n_items,1))





from cvx import *
from misc import *
from num import *
from kernels import *
from binarizer import *
from bool_kernels import *
from prop_kernels import *

__all__ = ["cvx", "misc", "num", "kernels", "bool_kernels", "prop_kernels", "binarizer"]